class ArangodanticError(Exception):
    """Generic Arangodantic error class."""

    pass


class ModelNotFoundError(ArangodanticError):
    pass


class UniqueConstraintError(ArangodanticError):
    pass


class GraphNotFoundError(ArangodanticError):
    pass


class ConfigError(ArangodanticError):
    pass



from asyncio import gather
from uuid import uuid4

import pytest

from arangodantic import GraphNotFoundError, ModelNotFoundError, UniqueConstraintError
from arangodantic.tests.conftest import (
    Person,
    Relation,
    RelationGraph,
    SecondaryRelation,
    SecondaryRelationGraph,
)


@pytest.mark.asyncio
async def test_save_through_graph(relation_graph):
    alice = Person(name="Alice")
    bob = Person(name="Bob")

    await RelationGraph.save(alice)
    assert alice.key_
    assert alice.rev_
    await RelationGraph.save(bob)

    alice.name = "Alice in Wonderland"
    await RelationGraph.save(alice)
    await alice.reload()
    assert alice.name == "Alice in Wonderland"

    ab = Relation(_from=alice, _to=bob, kind="BFF")
    await RelationGraph.save(ab)
    assert ab.key_
    assert ab.rev_
    ab.kind = "BF"
    await RelationGraph.save(ab)
    await ab.reload()
    assert ab.kind == "BF"


@pytest.mark.asyncio
async def test_unique_constraint_graph(relation_graph):
    # Create unique index on the "name" field.
    await Person.get_collection().add_hash_index(fields=["name"], unique=True)

    pre_generated_key = str(uuid4())

    person = Person(name="John Doe", _key=pre_generated_key)
    await RelationGraph.save(person)

    assert person.rev_ is not None

    loaded_person = await Person.load(pre_generated_key)
    assert loaded_person.key_ == pre_generated_key

    # Colliding primary key
    person_2 = Person(name="Jane Doe", _key=pre_generated_key)
    with pytest.raises(UniqueConstraintError):
        await RelationGraph.save(person_2)

    person_2.key_ = None
    await person_2.save()

    # Colliding "name"
    person_3 = Person(name="Jane Doe")
    with pytest.raises(UniqueConstraintError):
        await RelationGraph.save(person_3)

    person_3.name = "Jane Jr. Doe"
    await RelationGraph.save(person_3)

    with pytest.raises(UniqueConstraintError):
        person_3.name = "Jane Doe"
        await RelationGraph.save(person_3)


@pytest.mark.asyncio
async def test_deletion_through_graph(relation_graph, secondary_relation_graph):
    # Create some example persons
    alice = Person(name="Alice")
    bob = Person(name="Bob")
    malory = Person(name="Malory")
    await gather(alice.save(), bob.save(), malory.save())

    # Create some edges between them in the primary graph
    ab = Relation(_from=alice, _to=bob, kind="BFF")
    await ab.save()
    am = Relation(_from=alice, _to=malory, kind="hates")
    await am.save()

    # Create some edges between them in the secondary graph
    ab2 = SecondaryRelation(_from=alice, _to=bob, kind="knows")
    await ab2.save()
    am2 = SecondaryRelation(_from=alice, _to=malory, kind="hates")
    await am2.save()

    # Delete document normally (not using graph) and check edges were left in place
    assert await malory.delete()
    await am.reload()
    await am2.reload()
    # Clean up orphaned edges manually
    assert await RelationGraph.delete(am)
    assert await SecondaryRelationGraph.delete(am2)
    assert not await RelationGraph.delete(am, ignore_missing=True)
    assert not await SecondaryRelationGraph.delete(am2, ignore_missing=True)

    # Delete document through graph and verify edges were deleted from both graphs.
    assert await RelationGraph.delete(bob)
    with pytest.raises(ModelNotFoundError):
        await ab.reload()
    with pytest.raises(ModelNotFoundError):
        await ab2.reload()
    with pytest.raises(ModelNotFoundError):
        await RelationGraph.delete(ab)
    with pytest.raises(ModelNotFoundError):
        await SecondaryRelationGraph.delete(ab2)
    assert not await RelationGraph.delete(ab, ignore_missing=True)
    assert not await SecondaryRelationGraph.delete(ab2, ignore_missing=True)

    assert not await RelationGraph.delete(bob, ignore_missing=True)
    with pytest.raises(ModelNotFoundError):
        await RelationGraph.delete(bob)


@pytest.mark.asyncio
async def test_delete_graph(relation_graph):
    assert await RelationGraph.delete_graph()
    with pytest.raises(GraphNotFoundError):
        assert not await RelationGraph.delete_graph()
    assert not await RelationGraph.delete_graph(ignore_missing=True)

    assert await RelationGraph.get_db().has_collection(Person.get_collection_name())
    assert await RelationGraph.get_db().has_collection(Relation.get_collection_name())

    await RelationGraph.ensure_graph()
    assert await RelationGraph.delete_graph(drop_collections=True)
    with pytest.raises(GraphNotFoundError):
        assert not await RelationGraph.delete_graph()
    assert not await RelationGraph.get_db().has_collection(Person.get_collection_name())
    assert not await RelationGraph.get_db().has_collection(
        Relation.get_collection_name()
    )



from typing import Callable, Optional

from aioarangodb.database import StandardDatabase
from inflection import pluralize, underscore
from pydantic import BaseModel


class Configuration(BaseModel):
    """
    Provide storage for global configurations.
    """

    db: Optional[StandardDatabase] = None
    prefix: str = ""
    key_gen: Optional[Callable] = None
    collection_generator: Optional[Callable] = None
    graph_generator: Optional[Callable] = None
    lock: Optional[Callable] = None
    lock_name_prefix = "arangodantic_"

    class Config:
        arbitrary_types_allowed = True


CONF = Configuration()


def pluralize_underscore_class(cls: Callable) -> str:
    """
    Pluralize and underscore the name of the class.

    :param cls: The class.
    :return: The pluralized and underscored name of the class.
    """
    return underscore(pluralize(cls.__name__))


def underscore_class(cls: Callable) -> str:
    """
    Underscore the name of the class.

    :param cls: The class.
    :return: The underscored name of the class.
    """
    return underscore(cls.__name__)


def configure(
    db: StandardDatabase,
    *,
    prefix: str = "",
    key_gen: Optional[Callable] = None,
    collection_generator: Optional[Callable] = pluralize_underscore_class,
    graph_generator: Optional[Callable] = underscore_class,
    lock: Optional[Callable] = None,
) -> None:
    """Configures the DB.

    :param db: The aioarangodb StandardDatabase.
    :param prefix: A prefix to use for collection names.
    :param key_gen: A function that generates new "_key"s.
    :param collection_generator: A function that converts the class to a collection
    name.
    :param graph_generator: A function that converts the class to a graph name.
    :param lock: A lock class (for example AsyncLock from Shylock) that provides access
    to named locks.
    """
    CONF.db = db
    CONF.prefix = prefix
    CONF.key_gen = key_gen
    CONF.collection_generator = collection_generator
    CONF.graph_generator = graph_generator
    CONF.lock = lock



# flake8: noqa
from arangodantic.configurations import CONF, configure
from arangodantic.exceptions import *
from arangodantic.graphs import ArangodanticGraphConfig, EdgeDefinition, Graph
from arangodantic.models import ArangodanticCollectionConfig, DocumentModel, EdgeModel



from uuid import uuid4

import pytest

from arangodantic import ModelNotFoundError, UniqueConstraintError
from arangodantic.tests.conftest import ExtendedIdentity, Identity, Link, SubModel


@pytest.mark.asyncio
async def test_save_and_load_model(identity_collection):
    identity = Identity(name="John Doe")
    await identity.save()

    assert identity.key_ is not None
    assert identity.rev_ is not None

    loaded_identity = await Identity.load(identity.key_)
    assert identity.key_ == loaded_identity.key_


@pytest.mark.asyncio
async def test_unique_constraint(identity_collection):
    # Create unique index on the "name" field.
    await Identity.get_collection().add_hash_index(fields=["name"], unique=True)

    pre_generated_key = str(uuid4())

    identity = Identity(name="John Doe", _key=pre_generated_key)
    await identity.save()

    assert identity.rev_ is not None

    loaded_identity = await Identity.load(pre_generated_key)
    assert loaded_identity.key_ == pre_generated_key

    # Colliding primary key
    identity_2 = Identity(name="Jane Doe", _key=pre_generated_key)
    with pytest.raises(UniqueConstraintError):
        await identity_2.save()

    identity_2.key_ = None
    await identity_2.save()

    # Colliding "name"
    identity_3 = Identity(name="Jane Doe")
    with pytest.raises(UniqueConstraintError):
        await identity_3.save()

    identity_3.name = "Jane Jr. Doe"
    await identity_3.save()

    with pytest.raises(UniqueConstraintError):
        identity_3.name = "Jane Doe"
        await identity_3.save()


@pytest.mark.asyncio
async def test_delete_model(identity_collection):
    identity = Identity(name="Jane Doe")
    await identity.save()
    assert await identity.delete() is True

    with pytest.raises(ModelNotFoundError):
        await identity.delete()

    assert await identity.delete(ignore_missing=True) is False


@pytest.mark.asyncio
async def test_reload(identity_collection):
    identity = Identity(name="Jane Doe")
    with pytest.raises(ModelNotFoundError):
        await identity.reload()

    await identity.save()

    loaded_identity = await Identity.load(identity.key_)
    assert identity.key_ == loaded_identity.key_

    identity.name = "Jane Austen"
    await identity.save()

    assert loaded_identity.name == "Jane Doe"
    await loaded_identity.reload()
    assert loaded_identity.name == "Jane Austen"


@pytest.mark.asyncio
async def test_locking(identity_collection):
    identity = Identity(name="James Doe")
    await identity.save()

    second_lock = Identity.get_lock(identity.key_)

    # Lock and reload the identity
    async with identity.lock_and_reload():
        # Verify the lock is held
        assert await second_lock.acquire(block=False) is False

    # Verify the lock is not held any longer
    try:
        assert await second_lock.acquire(block=False) is True
    finally:
        await second_lock.release()

    # Lock and load as new identity
    async with Identity.lock_and_load(identity.key_) as _:
        # Verify the lock is held
        assert await second_lock.acquire(block=False) is False

    # Verify the lock is not held any longer
    try:
        assert await second_lock.acquire(block=False) is True
    finally:
        await second_lock.release()


@pytest.mark.asyncio
async def test__before_save(extended_identity_collection):
    identity = ExtendedIdentity(name="John Doe")
    await identity.save(extra="foo")
    identity = await ExtendedIdentity.load(identity.key_)
    assert identity.extra == "foo"


@pytest.mark.asyncio
async def test_sub_models(extended_identity_collection):
    sub = SubModel(text="foo")
    identity = ExtendedIdentity(name="John Doe", sub=sub)
    await identity.save()

    identity = await ExtendedIdentity.load(identity.key_)
    assert isinstance(identity.sub, SubModel)
    assert identity.sub.text == "foo"

    identity.sub = None
    await identity.reload()
    assert isinstance(identity.sub, SubModel)
    assert identity.sub.text == "foo"


@pytest.mark.asyncio
async def test_edge_model(identity_collection, link_collection):
    alice = Identity(name="Alice")
    await alice.save()
    bob = Identity(name="Bob")
    await bob.save()

    link = Link(_from=alice, _to=bob, type="Knows")
    await link.save()

    await link.reload()
    assert link.from_ == alice.id_
    assert link.to_ == bob.id_



# Partial backport of contextlib from python 3.7
# https://github.com/python/cpython/blob/3.7/Lib/contextlib.py
# and _check_methods from
# https://github.com/python/cpython/blob/3.7/Lib/_collections_abc.py
#
# This code is licensed under PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2
# https://github.com/python/cpython/blob/3.7/LICENSE

"""Utilities for with-statement contexts.  See PEP 343."""
import abc
import sys
from functools import wraps


def _check_methods(C, *methods):
    mro = C.__mro__
    for method in methods:
        for B in mro:
            if method in B.__dict__:
                if B.__dict__[method] is None:
                    return NotImplemented
                break
        else:
            return NotImplemented
    return True


class AbstractContextManager(abc.ABC):

    """An abstract base class for context managers."""

    def __enter__(self):
        """Return `self` upon entering the runtime context."""
        return self

    @abc.abstractmethod
    def __exit__(self, exc_type, exc_value, traceback):
        """Raise any exception triggered within the runtime context."""
        return None

    @classmethod
    def __subclasshook__(cls, C):
        if cls is AbstractContextManager:
            return _check_methods(C, "__enter__", "__exit__")
        return NotImplemented


class AbstractAsyncContextManager(abc.ABC):

    """An abstract base class for asynchronous context managers."""

    async def __aenter__(self):
        """Return `self` upon entering the runtime context."""
        return self

    @abc.abstractmethod
    async def __aexit__(self, exc_type, exc_value, traceback):
        """Raise any exception triggered within the runtime context."""
        return None

    @classmethod
    def __subclasshook__(cls, C):
        if cls is AbstractAsyncContextManager:
            return _check_methods(C, "__aenter__", "__aexit__")
        return NotImplemented


class ContextDecorator(object):
    "A base class or mixin that enables context managers to work as decorators."

    def _recreate_cm(self):
        """Return a recreated instance of self.
        Allows an otherwise one-shot context manager like
        _GeneratorContextManager to support use as
        a decorator via implicit recreation.
        This is a private interface just for _GeneratorContextManager.
        See issue #11647 for details.
        """
        return self

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwds):
            with self._recreate_cm():
                return func(*args, **kwds)

        return inner


class _GeneratorContextManagerBase:
    """Shared functionality for @contextmanager and @asynccontextmanager."""

    def __init__(self, func, args, kwds):
        self.gen = func(*args, **kwds)
        self.func, self.args, self.kwds = func, args, kwds
        # Issue 19330: ensure context manager instances have good docstrings
        doc = getattr(func, "__doc__", None)
        if doc is None:
            doc = type(self).__doc__
        self.__doc__ = doc
        # Unfortunately, this still doesn't provide good help output when
        # inspecting the created context manager instances, since pydoc
        # currently bypasses the instance docstring and shows the docstring
        # for the class instead.
        # See http://bugs.python.org/issue19404 for more details.


class _GeneratorContextManager(
    _GeneratorContextManagerBase, AbstractContextManager, ContextDecorator
):
    """Helper for @contextmanager decorator."""

    def _recreate_cm(self):
        # _GCM instances are one-shot context managers, so the
        # CM must be recreated each time a decorated function is
        # called
        return self.__class__(self.func, self.args, self.kwds)

    def __enter__(self):
        # do not keep args and kwds alive unnecessarily
        # they are only needed for recreation, which is not possible anymore
        del self.args, self.kwds, self.func
        try:
            return next(self.gen)
        except StopIteration:
            raise RuntimeError("generator didn't yield") from None

    def __exit__(self, type, value, traceback):
        if type is None:
            try:
                next(self.gen)
            except StopIteration:
                return False
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                # Need to force instantiation so we can reliably
                # tell if we get the same exception back
                value = type()
            try:
                self.gen.throw(type, value, traceback)
            except StopIteration as exc:
                # Suppress StopIteration *unless* it's the same exception that
                # was passed to throw().  This prevents a StopIteration
                # raised inside the "with" statement from being suppressed.
                return exc is not value
            except RuntimeError as exc:
                # Don't re-raise the passed in exception. (issue27122)
                if exc is value:
                    return False
                # Likewise, avoid suppressing if a StopIteration exception
                # was passed to throw() and later wrapped into a RuntimeError
                # (see PEP 479).
                if type is StopIteration and exc.__cause__ is value:
                    return False
                raise
            except:  # noqa
                # only re-raise if it's *not* the exception that was
                # passed to throw(), because __exit__() must not raise
                # an exception unless __exit__() itself failed.  But throw()
                # has to raise the exception to signal propagation, so this
                # fixes the impedance mismatch between the throw() protocol
                # and the __exit__() protocol.
                #
                # This cannot use 'except BaseException as exc' (as in the
                # async implementation) to maintain compatibility with
                # Python 2, where old-style class exceptions are not caught
                # by 'except BaseException'.
                if sys.exc_info()[1] is value:
                    return False
                raise
            raise RuntimeError("generator didn't stop after throw()")


class _AsyncGeneratorContextManager(
    _GeneratorContextManagerBase, AbstractAsyncContextManager
):
    """Helper for @asynccontextmanager."""

    async def __aenter__(self):
        try:
            return await self.gen.__anext__()
        except StopAsyncIteration:
            raise RuntimeError("generator didn't yield") from None

    async def __aexit__(self, typ, value, traceback):
        if typ is None:
            try:
                await self.gen.__anext__()
            except StopAsyncIteration:
                return
            else:
                raise RuntimeError("generator didn't stop")
        else:
            if value is None:
                value = typ()
            # See _GeneratorContextManager.__exit__ for comments on subtleties
            # in this implementation
            try:
                await self.gen.athrow(typ, value, traceback)
                raise RuntimeError("generator didn't stop after athrow()")
            except StopAsyncIteration as exc:
                return exc is not value
            except RuntimeError as exc:
                if exc is value:
                    return False
                # Avoid suppressing if a StopIteration exception
                # was passed to throw() and later wrapped into a RuntimeError
                # (see PEP 479 for sync generators; async generators also
                # have this behavior). But do this only if the exception wrapped
                # by the RuntimeError is actully Stop(Async)Iteration (see
                # issue29692).
                if isinstance(value, (StopIteration, StopAsyncIteration)):
                    if exc.__cause__ is value:
                        return False
                raise
            except BaseException as exc:
                if exc is not value:
                    raise


def asynccontextmanager(func):
    """@asynccontextmanager decorator.
    Typical usage:
        @asynccontextmanager
        async def some_async_generator(<arguments>):
            <setup>
            try:
                yield <value>
            finally:
                <cleanup>
    This makes this:
        async with some_async_generator(<arguments>) as <variable>:
            <body>
    equivalent to this:
        <setup>
        try:
            <variable> = <value>
            <body>
        finally:
            <cleanup>
    """

    @wraps(func)
    def helper(*args, **kwds):
        return _AsyncGeneratorContextManager(func, args, kwds)

    return helper



from abc import ABC
from functools import lru_cache
from typing import Dict, List, Optional, Type, Union

import aioarangodb.graph
import pydantic
from aioarangodb import GraphDeleteError
from aioarangodb.database import StandardDatabase
from pydantic import Field

from arangodantic import GraphNotFoundError, ModelNotFoundError, UniqueConstraintError
from arangodantic.arangdb_error_codes import (
    ERROR_ARANGO_DOCUMENT_NOT_FOUND,
    ERROR_ARANGO_UNIQUE_CONSTRAINT_VIOLATED,
    ERROR_GRAPH_NOT_FOUND,
)
from arangodantic.configurations import CONF
from arangodantic.models import DocumentModel, EdgeModel, Model


class EdgeDefinition(pydantic.BaseModel):
    edge_collection: Type[EdgeModel]
    from_vertex_collections: List[Type[DocumentModel]]
    to_vertex_collections: List[Type[DocumentModel]]


class ArangodanticGraphConfig(pydantic.BaseModel):
    graph_name: Optional[str] = Field(
        None, description="Override the name of the graph to use"
    )
    edge_definitions: Optional[List[EdgeDefinition]] = None
    orphan_collections: Optional[List[Model]] = None


class Graph(ABC):
    @classmethod
    @lru_cache()
    def get_graph_name(cls) -> str:
        cls_config: ArangodanticGraphConfig = getattr(
            cls, "ArangodanticConfig", ArangodanticGraphConfig()
        )
        if getattr(cls_config, "graph_name", None):
            graph = cls_config.graph_name
        else:
            graph = CONF.graph_generator(cls)  # type: ignore
        return f"{CONF.prefix}{graph}"

    @classmethod
    def get_db(cls) -> StandardDatabase:
        return CONF.db

    @classmethod
    def get_graph(cls) -> aioarangodb.graph.Graph:
        return cls.get_db().graph(cls.get_graph_name())

    @classmethod
    async def save(cls, model: Union[DocumentModel, EdgeModel], **kwargs):
        """
        Save the model through the graph; either creates a new one or updates/replaces
        an existing document.

        Doing this through the graph will make ArangoDB enforce some better validation.
        For more details please see
        https://www.arangodb.com/docs/stable/graphs.html#manipulating-collections-of-named-graphs-with-regular-document-functions

        :raise UniqueConstraintViolated: Raised when there is a unique constraint
        violation.
        """

        graph = cls.get_graph()

        if not model.rev_:
            # Insert new document
            if not model.key_ and CONF.key_gen:
                # Use generator to generate new key
                model.key_ = str(CONF.key_gen())

            await model.before_save(new=True, **kwargs)

            data = model.get_arangodb_data()
            if not model.key_:
                # Let ArangoDB handle key generation
                del data["_key"]

            try:
                collection_name = model.get_collection_name()
                if isinstance(model, EdgeModel):
                    response = await graph.insert_edge(
                        collection=collection_name, edge=data
                    )
                else:
                    response = await graph.insert_vertex(
                        collection=collection_name, vertex=data
                    )
            except aioarangodb.exceptions.DocumentInsertError as ex:
                if ex.error_code == ERROR_ARANGO_UNIQUE_CONSTRAINT_VIOLATED:
                    raise UniqueConstraintError(ex.error_message)
                raise
        else:
            # Update existing document
            await model.before_save(new=False, **kwargs)
            data = model.get_arangodb_data()
            try:
                if isinstance(model, EdgeModel):
                    response = await graph.replace_edge(edge=data)
                else:
                    response = await graph.replace_vertex(vertex=data)
            except aioarangodb.exceptions.DocumentReplaceError as ex:
                if ex.error_code == ERROR_ARANGO_UNIQUE_CONSTRAINT_VIOLATED:
                    raise UniqueConstraintError(ex.error_message)
                raise

        model.key_ = response["_key"]
        model.rev_ = response["_rev"]

    @classmethod
    async def delete_vertex(cls, document: DocumentModel, ignore_missing=False) -> bool:
        """
        Delete a vertex from the graph. This will also ensure all edges to/from the
        vertex are deleted.

        :param document: The document model to delete.
        :param ignore_missing: Do not raise an exception on missing document.
        :raise ModelNotFoundError: Raised if the document is not found in the database
        and ignore_missing is set to False (the default value).
        :return: True if document was deleted successfully, False if document was
        not found and **ignore_missing** was set to True.
        """
        data = document.get_arangodb_data()
        try:
            result: bool = await cls.get_graph().delete_vertex(
                data, ignore_missing=ignore_missing
            )
        except aioarangodb.exceptions.DocumentDeleteError as ex:
            if ex.error_code == ERROR_ARANGO_DOCUMENT_NOT_FOUND:
                raise ModelNotFoundError(
                    f"No '{document.__class__.__name__}' found with _key "
                    f"'{document.key_}'"
                )
            raise

        return result

    @classmethod
    async def delete_edge(cls, edge: EdgeModel, ignore_missing=False) -> bool:
        """
        Delete an edge from the graph.

        :param edge: The edge model to delete.
        :param ignore_missing: Do not raise an exception on missing document.
        :raise ModelNotFoundError: Raised if the document is not found in the database
        and ignore_missing is set to False (the default value).
        :return: True if edge was deleted successfully, False if edge was
        not found and **ignore_missing** was set to True.
        """
        data = edge.get_arangodb_data()
        try:
            result: bool = await cls.get_graph().delete_edge(
                data, ignore_missing=ignore_missing
            )
        except aioarangodb.exceptions.DocumentDeleteError as ex:
            if ex.error_code == ERROR_ARANGO_DOCUMENT_NOT_FOUND:
                raise ModelNotFoundError(
                    f"No '{edge.__class__.__name__}' found with _key '{edge.key_}'"
                )
            raise

        return result

    @classmethod
    async def delete(
        cls, model: Union[DocumentModel, EdgeModel], ignore_missing=False
    ) -> bool:
        """
        Delete a model (edge or vertex) from the graph. This will also ensure all edges
        to/from the vertex are deleted.

        :param model: The model to delete.
        :param ignore_missing: Do not raise an exception on missing document.
        :raise ModelNotFoundError: Raised if the document is not found in the database
        and ignore_missing is set to False (the default value).
        :return: True if model was deleted successfully, False if model was
        not found and **ignore_missing** was set to True.
        """
        if isinstance(model, EdgeModel):
            return await cls.delete_edge(edge=model, ignore_missing=ignore_missing)
        else:
            return await cls.delete_vertex(
                document=model, ignore_missing=ignore_missing
            )

    @classmethod
    async def ensure_graph(cls, **kwargs):
        """
        Ensure the graph exists and create it if needed.
        """
        cls_config: ArangodanticGraphConfig = getattr(
            cls, "ArangodanticConfig", ArangodanticGraphConfig()
        )

        def get_edge_definitions() -> List[Dict[str, Union[str, List[str]]]]:
            edge_definitions = getattr(cls_config, "edge_definitions", None)
            if not edge_definitions:
                edge_definitions = []
            return [
                {
                    "edge_collection": ed.edge_collection.get_collection_name(),
                    "from_vertex_collections": [
                        vc.get_collection_name() for vc in ed.from_vertex_collections
                    ],
                    "to_vertex_collections": [
                        vc.get_collection_name() for vc in ed.to_vertex_collections
                    ],
                }
                for ed in edge_definitions
            ]

        def get_orphan_collections() -> List[str]:
            orphan_collections = getattr(cls_config, "orphan_collections", None)
            if not orphan_collections:
                orphan_collections = []
            return [oc.get_collection_name() for oc in orphan_collections]

        name = cls.get_graph_name()
        db = cls.get_db()

        if not await db.has_graph(name):
            await db.create_graph(
                name,
                edge_definitions=get_edge_definitions(),
                orphan_collections=get_orphan_collections(),
                **kwargs,
            )

    @classmethod
    async def delete_graph(cls, ignore_missing: bool = False, drop_collections=None):
        """
        Delete the graph if it exists.

        :param ignore_missing: Do not raise an exception on missing graph.
        :param drop_collections: Drop the collections of the graph also. This
        is only if they are not in use by other graphs.
        :raise GraphNotFoundError: If the graph does not exist and **ignore_missing** is
        False
        """
        name = cls.get_graph_name()
        db = cls.get_db()

        try:
            return await db.delete_graph(
                name, ignore_missing=ignore_missing, drop_collections=drop_collections
            )
        except GraphDeleteError as ex:
            if ex.error_code == ERROR_GRAPH_NOT_FOUND:
                raise GraphNotFoundError(f"No graph found with name '{name}'")
            raise



import random
import string
from typing import Optional
from uuid import uuid4

import pydantic
import pytest
from aioarangodb import ArangoClient
from shylock import AsyncLock as Lock
from shylock import ShylockAioArangoDBBackend
from shylock import configure as configure_shylock

from arangodantic import DocumentModel, EdgeDefinition, EdgeModel, Graph, configure

HOSTS = "http://localhost:8529"
USERNAME = "root"
PASSWORD = ""
DATABASE = "test"


@pytest.fixture
async def configure_db():
    def rand_str(length: int) -> str:
        """
        Generate a random string for collection names.

        :param length: The length of the random string.
        :return: The random prefix string.
        """
        chars = string.ascii_letters + string.digits
        return "".join(random.choice(chars) for _ in range(length))

    prefix = f"test-{rand_str(10)}"

    client = ArangoClient(hosts=HOSTS)
    # Connect to "_system" database and create the actual database if it doesn't exist
    sys_db = await client.db("_system", username=USERNAME, password=PASSWORD)
    if not await sys_db.has_database(DATABASE):
        await sys_db.create_database(DATABASE)

    db = await client.db(DATABASE, username=USERNAME, password=PASSWORD)
    configure_shylock(await ShylockAioArangoDBBackend.create(db, f"{prefix}-shylock"))
    configure(db, prefix=f"{prefix}-", key_gen=uuid4, lock=Lock)

    yield

    await db.delete_collection(f"{prefix}-shylock")
    await client.close()


class Identity(DocumentModel):
    """Dummy identity Arangodantic model."""

    name: str = ""


class SubModel(pydantic.BaseModel):
    """Dummy plain pydantic sub-model."""

    text: str = ""


class ExtendedIdentity(Identity):
    """Dummy extended identity Arangodantic model."""

    extra: Optional[str] = None
    sub: Optional[SubModel] = None

    class ArangodanticConfig:
        collection_name = "ext_identities"

    async def before_save(self, new: bool, extra: Optional[str] = None, **kwargs):
        self.extra = extra


class Link(EdgeModel):
    """Dummy Arangodantic edge model."""

    type: str


class Person(DocumentModel):
    """Documents describing persons."""

    name: str


class Relation(EdgeModel):
    """Edge documents describing relation between people."""

    kind: str


class SecondaryRelation(EdgeModel):
    """Edge documents describing a secondary relation between people."""

    kind: str


class RelationGraph(Graph):
    class ArangodanticConfig:
        edge_definitions = [
            EdgeDefinition(
                edge_collection=Relation,
                from_vertex_collections=[Person],
                to_vertex_collections=[Person],
            )
        ]


class SecondaryRelationGraph(Graph):
    class ArangodanticConfig:
        edge_definitions = [
            EdgeDefinition(
                edge_collection=SecondaryRelation,
                from_vertex_collections=[Person],
                to_vertex_collections=[Person],
            )
        ]


@pytest.fixture
async def identity_collection(configure_db):
    await Identity.ensure_collection()
    yield
    await Identity.delete_collection()


@pytest.fixture
async def extended_identity_collection(configure_db):
    await ExtendedIdentity.ensure_collection()
    yield
    await ExtendedIdentity.delete_collection()


@pytest.fixture
async def link_collection(configure_db):
    await Link.ensure_collection()
    yield
    await Link.delete_collection()


@pytest.fixture
async def relation_graph(configure_db):
    await RelationGraph.ensure_graph()
    yield
    await RelationGraph.delete_graph(ignore_missing=True, drop_collections=True)


@pytest.fixture
async def secondary_relation_graph(configure_db):
    await SecondaryRelationGraph.ensure_graph()
    yield
    await SecondaryRelationGraph.delete_graph(
        ignore_missing=True, drop_collections=True
    )



from abc import ABC
from functools import lru_cache
from typing import Optional, Type, TypeVar, Union

import aioarangodb.exceptions
import pydantic
from aioarangodb.collection import StandardCollection
from aioarangodb.database import StandardDatabase
from pydantic import Field

from arangodantic.arangdb_error_codes import (
    ERROR_ARANGO_DOCUMENT_NOT_FOUND,
    ERROR_ARANGO_UNIQUE_CONSTRAINT_VIOLATED,
)
from arangodantic.configurations import CONF
from arangodantic.exceptions import (
    ConfigError,
    ModelNotFoundError,
    UniqueConstraintError,
)

try:
    from contextlib import asynccontextmanager  # type: ignore
except ImportError:
    from arangodantic.asynccontextmanager import asynccontextmanager

TModel = TypeVar("TModel", bound="Model")


class ArangodanticCollectionConfig(pydantic.BaseModel):
    collection_name: Optional[str] = Field(
        None, description="Override the name of the collection to use"
    )


class Model(pydantic.BaseModel, ABC):
    """
    Base model class.

    Implements basic functionality for Pydantic based models, such as save, delete, get
    etc.
    """

    key_: Optional[str] = Field(alias="_key")
    rev_: Optional[str] = Field(alias="_rev")

    @property
    def id_(self) -> Optional[str]:
        """
        The "_id" field used by ArangoDB (if set).
        """
        if self.key_ is None:
            return None
        return f"{self.get_collection_name()}/{self.key_}"

    @classmethod
    def get_lock_name(cls, key: str) -> str:
        """
        Get the name for the lock for a particular key.
        """
        return f"{CONF.lock_name_prefix}{cls.get_collection_name()}_{key}"

    @classmethod
    def get_lock(cls, key: str):
        """
        Get a named lock corresponding to an instance of this class with a given key.

        :raise ConfigError: Raised if lock is not configured.
        """
        lock_name = cls.get_lock_name(key)
        if not CONF.lock:
            raise ConfigError("Trying to get lock when no lock is configured")
        return CONF.lock(lock_name)

    @classmethod
    async def load(cls: Type[TModel], key: str) -> TModel:
        """
        Get a model based on the ArangoDB "_key".

        :param key: The "_key" of the model to get.
        :return: The model.
        :raise ModelNotFoundError: Raised if no matching document is found.
        """
        response = await cls.get_collection().get(document={"_key": key})
        if response is None:
            raise ModelNotFoundError(f"No '{cls.__name__}' found with _key '{key}'")

        return cls(**response)

    async def reload(self) -> None:
        """
        Reload the model from the database.

        :raise ModelNotFoundError: Raised if the document is not found in the database.
        """
        if not self.key_:
            raise ModelNotFoundError(
                f"Can't reload '{self.__class__.__name__}' without a key"
            )
        new = await self.load(self.key_)
        self.__dict__.update(new.__dict__)

    @classmethod
    @asynccontextmanager
    async def lock_and_load(cls, key):
        """
        Context manager to acquire a lock for a model, load it and return it and upon
        exit release the lock.
        """
        async with cls.get_lock(key):
            yield await cls.load(key)

    @asynccontextmanager
    async def lock_and_reload(self):
        """
        Context manager to lock the current model and reload it from the database.
        Upon exit release the lock.
        """
        async with self.get_lock(self.key_):
            await self.reload()
            yield

    async def save(self, **kwargs) -> None:
        """
        Save the document; either creates a new one or updates/replaces an existing
        document.

        :raise UniqueConstraintViolated: Raised when there is a unique constraint
        violation.
        """

        if not self.rev_:
            # Insert new document
            if not self.key_ and CONF.key_gen:
                # Use generator to generate new key
                self.key_ = str(CONF.key_gen())

            await self.before_save(new=True, **kwargs)

            data = self.get_arangodb_data()
            if not self.key_:
                # Let ArangoDB handle key generation
                del data["_key"]

            try:
                response = await self.get_collection().insert(document=data)
            except aioarangodb.exceptions.DocumentInsertError as ex:
                if ex.error_code == ERROR_ARANGO_UNIQUE_CONSTRAINT_VIOLATED:
                    raise UniqueConstraintError(ex.error_message)
                raise
        else:
            # Update existing document
            await self.before_save(new=False, **kwargs)
            data = self.get_arangodb_data()
            try:
                response = await self.get_collection().replace(document=data)
            except aioarangodb.exceptions.DocumentReplaceError as ex:
                if ex.error_code == ERROR_ARANGO_UNIQUE_CONSTRAINT_VIOLATED:
                    raise UniqueConstraintError(ex.error_message)
                raise

        self.key_ = response["_key"]
        self.rev_ = response["_rev"]

    async def delete(self, ignore_missing=False) -> bool:
        """
        Delete the document.

        :param ignore_missing: Do not raise an exception on missing document.
        :return: Returns False if document was not found and ignore_missing was set to
        True, else always returns True.
        :raise ModelNotFoundError: Raised if the document is not found in the database
        and ignore_missing is set to False (the default value).
        """

        data = self.get_arangodb_data()
        try:
            result: bool = await self.get_collection().delete(
                document=data, silent=True, ignore_missing=ignore_missing
            )
        except aioarangodb.exceptions.DocumentDeleteError as ex:
            if ex.error_code == ERROR_ARANGO_DOCUMENT_NOT_FOUND:
                raise ModelNotFoundError(
                    f"No '{self.__class__.__name__}' found with _key '{self.key_}'"
                )
            raise

        return result

    def get_arangodb_data(self) -> dict:
        """
        Get a dictionary of the data to pass on to ArangoDB when inserting, updating and
        deleting the document.
        """
        data = self.dict(by_alias=True)
        data["_id"] = self.id_
        return data

    @classmethod
    def get_db(cls) -> StandardDatabase:
        return CONF.db

    @classmethod
    @lru_cache()
    def get_collection_name(cls) -> str:
        cls_config: ArangodanticCollectionConfig = getattr(
            cls, "ArangodanticConfig", ArangodanticCollectionConfig()
        )
        if getattr(cls_config, "collection_name", None):
            collection = cls_config.collection_name
        else:
            collection = CONF.collection_generator(cls)  # type: ignore
        return f"{CONF.prefix}{collection}"

    @classmethod
    def get_collection(cls) -> StandardCollection:
        return cls.get_db().collection(cls.get_collection_name())

    @classmethod
    async def ensure_collection(cls, *args, **kwargs):
        """
        Ensure the collection exists and create it if needed.
        """
        name = cls.get_collection_name()
        db = cls.get_db()

        if not await db.has_collection(name):
            await db.create_collection(name, *args, **kwargs)

    @classmethod
    async def delete_collection(cls, ignore_missing: bool = True):
        """
        Delete the collection if it exists.

        :param ignore_missing: Do not raise an exception on missing collection.
        """
        name = cls.get_collection_name()
        db = cls.get_db()

        return await db.delete_collection(
            name, ignore_missing=ignore_missing, system=False
        )

    async def before_save(self, new: bool, **kwargs) -> None:
        """
        Function that's run before saving, should be overridden in subclasses if needed.

        :param new: Tells if the model is new (will be saved for the first time) or not.
        """
        pass


class DocumentModel(Model, ABC):
    """
    Base document model class.
    """


class EdgeModel(Model, ABC):
    """
    Base edge model class.
    """

    from_: Union[str, DocumentModel] = Field(alias="_from")
    to_: Union[str, DocumentModel] = Field(alias="_to")

    def get_arangodb_data(self) -> dict:
        """
        Get a dictionary of the data to pass on to ArangoDB when inserting, updating and
        deleting the document.
        """
        data = self.dict(by_alias=True, exclude={"from_", "to_"})
        data["_id"] = self.id_
        if isinstance(self.from_, DocumentModel):
            data["_from"] = self.from_.id_
        else:
            data["_from"] = self.from_

        if isinstance(self.to_, DocumentModel):
            data["_to"] = self.to_.id_
        else:
            data["_to"] = self.to_

        return data

    @classmethod
    async def ensure_collection(cls, *args, **kwargs):
        """
        Ensure the collection exists and create it if needed.
        """
        return await super(EdgeModel, cls).ensure_collection(edge=True, *args, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['arangodantic', 'arangodantic.tests']

package_data = \
{'': ['*']}

install_requires = \
['aioarangodb>=0.1.2,<0.2.0',
 'inflection>=0.5.1,<0.6.0',
 'pydantic>=1.6.1,<2.0.0']

extras_require = \
{'shylock': ['shylock[aioarangodb]>=1.1.1,<2.0.0']}

setup_kwargs = {
    'name': 'arangodantic',
    'version': '0.0.4',
    'description': 'Database models for ArangoDB using Pydantic base models.',
    'long_description': '# Arangodantic\n\n[![Build Status](https://travis-ci.com/digitalliving/arangodantic.svg?branch=master)](https://travis-ci.com/digitalliving/arangodantic)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![PyPI](https://img.shields.io/pypi/v/arangodantic)](https://pypi.org/project/arangodantic/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/arangodantic)](https://pypi.org/project/arangodantic/)\n[![License: BSD 3-Clause](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n\nDatabase models for ArangoDB using Pydantic base models.\n\n## Installation\n\nThe package is available on PyPi:\n\n```bash\npip install arangodantic\n\n# Or with Shylock\npip install arangodantic[shylock]\n```\n\n## Usage\n\nDefine your database models by extending either `DocumentModel` or `EdgeModel`.\nNested structures can be created by extending pydantic.BaseModel.\n\nConfigure Arangodantic. You can optionally define a collection name prefix,\na key generation function and a lock (needed if you want to use the locking\nfunctionality; [Shylock](https://github.com/lietu/shylock) is supported out of\nthe box, any other locking service such as\n[Sherlock](https://pypi.org/project/sherlock/) should at least in theory also\nwork).\n\nEnsure you have an ArangoDB server available with known credentials\n```bash\ndocker run --rm -p 8529:8529 -e ARANGO_ROOT_PASSWORD="" arangodb/arangodb:3.7.2.1\n```\n\n```python\nimport asyncio\nfrom uuid import uuid4\n\nfrom aioarangodb import ArangoClient\nfrom pydantic import BaseModel\nfrom shylock import AsyncLock as Lock\nfrom shylock import ShylockAioArangoDBBackend\nfrom shylock import configure as configure_shylock\n\nfrom arangodantic import DocumentModel, EdgeModel, configure\n\n\n# Define models\nclass Owner(BaseModel):\n    """Dummy owner Pydantic model."""\n\n    first_name: str\n    last_name: str\n\n\nclass Company(DocumentModel):\n    """Dummy company Arangodantic model."""\n\n    company_id: str\n    owner: Owner\n\n\nclass Link(EdgeModel):\n    """Dummy Link Arangodantic model."""\n\n    type: str\n\n\nasync def main():\n    # Configure the database settings\n    hosts = "http://localhost:8529"\n    username = "root"\n    password = ""\n    database = "example"\n    prefix = "example-"\n\n    client = ArangoClient(hosts=hosts)\n    # Connect to "_system" database and create the actual database if it doesn\'t exist\n    # Only for demo, you likely want to create the database in advance.\n    sys_db = await client.db("_system", username=username, password=password)\n    if not await sys_db.has_database(database):\n        await sys_db.create_database(database)\n\n    # Configure Arangodantic and Shylock\n    db = await client.db(database, username=username, password=password)\n    configure_shylock(await ShylockAioArangoDBBackend.create(db, f"{prefix}shylock"))\n    configure(db, prefix=prefix, key_gen=uuid4, lock=Lock)\n\n    # Create collections if they don\'t yet exist\n    # Only for demo, you likely want to create the collections in advance.\n    await Company.ensure_collection()\n    await Link.ensure_collection()\n\n    # Let\'s create some example entries\n    owner = Owner(first_name="John", last_name="Doe")\n    company = Company(company_id="1234567-8", owner=owner)\n    await company.save()\n    print(f"Company saved with key: {company.key_}")\n\n    second_owner = Owner(first_name="Jane", last_name="Doe")\n    second_company = Company(company_id="2345678-9", owner=second_owner)\n    await second_company.save()\n    print(f"Second company saved with key: {second_company.key_}")\n\n    link = Link(_from=company, _to=second_company, type="CustomerOf")\n    await link.save()\n    print(f"Link saved with key: {link.key_}")\n\n    # Hold named locks while loading and doing changes\n    async with Company.lock_and_load(company.key_) as c:\n        assert c.owner == owner\n        c.owner = second_owner\n        await c.save()\n\n    await company.reload()\n    assert c.owner == company.owner\n    print(f"Updated owner of company to \'{company.owner!r}\'")\n\n\nif __name__ == "__main__":\n    # Starting from Python 3.7 ->\n    # asyncio.run(main())\n\n    # Compatible with Python 3.6 ->\n    loop = asyncio.get_event_loop()\n    result = loop.run_until_complete(main())\n```\n\nYou might find [migrate-anything](https://github.com/cocreators-ee/migrate-anything) useful for creating and managing collections and indexes.\n\n## More examples\n- The [graph example](examples/graph_example.py) shows how arangodantic can be\n  used with graphs. Please note that graph related functionality is at the\n  moment really limited and will likely be extended later and might even be\n  restructured, so use with caution.\n\n## License\n\nThis code is released under the BSD 3-Clause license. Details in the\n[LICENSE](./LICENSE) file.\n',
    'author': 'Digital Living International Ltd',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/digitalliving/arangodantic',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
